// bcell.hh - Memory Cell Abstraction Interface (Bcell: Byte Mcell)
// GPLed - Stephan T. Lavavej - http://stl.caltech.edu

// The Bcell is derived from the Mcell<unsigned char> and overrides
// few of its methods. It stores no extra data. It simply
// provides extra methods which work on byte-oriented data.
// (The Bcell ruthlessly assumes that bytes are 8-bit.)

// Like Mcells, empty Bcells can be constructed, or given a name and
// a size upon construction. (Unlike Mcells, Bcells are zero-filled
// upon construction.) There is also a Bcell constructor that
// reads in a named file into memory. Finally, there is a Bcell copy
// constructor which allows Bcells and Mcell<unsigned char> to be
// used nearly interchangably. (As of yet there is no builtin method
// to downconvert a Bcell into an Mcell<unsigned char>, as a cast can
// do this easily.)

// With the print() method, the entire contents of a Bcell can be examined.

// Bcells support bitwise reading and writing in a big-endian fashion,
// though it is not quite as convenient as an Mcell<bool>.
// Bcells also support reading unsigned long (4 byte) and
// unsigned long long (8 byte) big-endian values.

// Bcells support many different types of compression. If ZIP is defined,
// the compress() and decompress(size_t originalsize) methods become
// available, which use zlib to compress the cell as much as possible.
// Also the bzip2() and unbzip2(unsigned int originalsize) methods become
// available, which use bzip2. Compression followed by decompression does
// not change the cell's contents at all.

// Bcells support Secure Hash Algorithm-1 one-way cryptographically secure
// hashing. The SHA1() method returns a 20-byte Bcell containing the SHA-1
// hash of our cell. SHA-512 has not yet been implemented in Bcell, but it
// easily could be.

// Bcells support easy file access. In addition to the file-reading
// constructor, the readFile(FILE *fp) method fills the cell with the
// entire contents of a file, and writeFile(FILE *fp) dumps the cell's
// contents to a file. writeFileName(const char *fname) does the same
// with a named file.

// Bcells support assignment, too. operator= which takes a
// const Mcell<unsigned char>& is provided so that Bcells work with all
// the usual Mcell operations. (+ and += return Mcells.) Also, a
// Bcell can be set with operator= to an unsigned long long int,
// which makes the Bcell hold 8 bytes which represent the unsigned
// long long int. (There does not yet exist a constructor for this, but
// writing one would be trivial.)
// NOTE: TAKE CARE TO CAST A VALUE TO UNSIGNED LONG LONG BEFORE ASSIGNING
// IT TO A BCELL. Eventually a method that works with unsigned longs, etc.
// may be provided.

// Bcells support frequency statistics: getFrequencies() returns an
// Mcell<size_t> whose Nth entry contains how many times byte N occurs
// in our cell.

// Finally, Bcells support the home-grown compression of bwtzip.
// bwtCompressStree() and bwtDecompressStree() use the suffix tree algorithm
// combined with several other improvements to compress our cell. Also
// provided are bwtCompressSarray() and bwtDecompressSarray() which use the
// suffix array algorithm.
// NOTE: THEY ARE NOT INTERCHANGABLE: IF YOU COMPRESS WITH SUFFIX ARRAYS YOU
// MUST USE THE SUFFIX ARRAY DECOMPRESSOR, AND LIKEWISE WITH SUFFIX TREES.
// This is due to minor differences in the algorithms, which otherwise achieve
// identical compression. Suffix trees are faster and linear-time.

// Behind the scenes, bwtzip compression is implemented with a host of
// functions private to Bcell. These are still written in a modular, fully
// abstracted fashion.

#ifndef BCELL_HH
#define BCELL_HH

#include <iostream>
#include <string>
#include <list>
#include <stack>
#include <utility>
#include <vector>
#include "common.hh"
#include "file.hh"
#include "mcell.hh"
#include "clock.hh"
#include "gcell.hh"
#include "ucell.hh"
#include "chunk.hh"

#ifdef ZIP
#include <zlib.h>
#endif
#ifdef BZIP
#include <bzlib.h>
#endif

using namespace std;

class Bcell : public Mcell<unsigned char> {

public:

// CONSTRUCTORS

    Bcell();                                    // Default constructor, creates an empty cell
    Bcell(string s, size_t n);                  // Construct a zero-filled cell
                                                // of a given name and size.
    Bcell(const char *fname);                   // Construct a cell whose contents are
                                                // that of a named file
    Bcell(const Mcell<unsigned char>& c);       // Copy constructor
    explicit Bcell(unsigned long long int ull); // Make our cell represent ull big-endian.
    Bcell(const Mcell<bool>& c);                // Copy constructor

// ACCESSORS, MUTATORS

    void setBit(size_t n, bool b);          // Set the nth bit of our cell to b

    bool getBit(size_t n) const;            // Get the nth bit of our cell

    unsigned long int                       // Return the unsigned long int
        readUL(size_t n) const;             // encoded big-endian by 4 bytes
                                            // starting at location n.
    unsigned long long int 
        readULL(size_t n) const;

    void print() const;                     // Print the contents of our cell to cout

// COMPRESSION

    Bcell& bwtCompressSarray(bool verbose = false);
    Bcell& bwtDecompressSarray(bool verbose = false);
    Bcell& bwtCompressStree(bool verbose = false);
    Bcell& bwtDecompressStree(bool verbose = false);

#ifdef ZIP
    Bcell& compress();                      // Compress our cell using zlib
    Bcell& decompress(size_t originalsize); // Decompress our cell using zlib
#endif
#ifdef BZIP
    Bcell& bzip2();                             // Compress our cell using bzip2
    Bcell& unbzip2(unsigned int originalsize);  // Decompress our Mcell using bzip2
#endif

// HASHING

    Bcell SHA1() const;                     // SHA-1 hash our memory cell
                                            // Returns 20-byte hash

// PNG

    unsigned long getCRC() const;
#ifdef ZIP
    Bcell& pngacc();
#endif
    Bcell& refilter(unsigned long width, unsigned long height,
        unsigned char bitdepth, unsigned char colortype, unsigned int filter);

// FILES

    Bcell& readFile(FILE *fp);              // Safely fill our cell with the
                                            // entire contents of a file.
                                            // Leave the file rewound and
                                            // return our cell.

    Bcell& writeFile(FILE *fp);             // Write the contents of our cell
                                            // to a file and return our cell.

    Bcell& writeFileName(                   // Write the contents of our
        const char *fname);                 // cell to a named file
                                            // and return our cell.
// ASSIGNMENT

    Bcell& operator=                        // Make our cell represent ull big-endian.
        (unsigned long long int ull);

    Bcell& operator=(const Mcell<bool>& c);

    Bcell& operator=(const Mcell<unsigned char>& c);
    Bcell& operator=(const list<unsigned char>& l);

// ETC

    Mcell<size_t> getFrequencies() const;   // Return an Mcell<size_t> of
                                            // length 256 with the byte
                                            // frequencies of our cell.

    Gcell<unsigned char> getGcell() const;

// DESTRUCTOR

    virtual ~Bcell();                       // Destructor

// FIXME: private should go here

// BWT FORWARD FUNCTIONS

    Bcell& bwtSarray();                     // Burrows-Wheeler Transform our cell, Suffix Array algorithm
    Bcell& bwtStree();                      // Burrows-Wheeler Transform
    Bcell& mtf();                           // MTF transform our cell
    Bcell& mtf2();                          // MTF-2 transform our cell
    Bcell& huff();                          // Huffman compress our cell
    Bcell& arith();                         // Adaptive arithmetic encode our cell
    Bcell& srle();                          // Switched-RLE encode our cell
    Bcell& srle0();                         // Switched-RLE-0 encode our cell
    Bcell& srle0a();                        // Switched-RLE-0-A encode our cell
    Bcell& zle();                           // Zero Length Encode our cell
    Bcell& zle1();                          // Zero Length Encode-1 our cell
    Bcell& zleWheeler();                    // Wheeler Zero Length Encode our cell
    Bcell& zleWheelerN(int n);              // Wheeler Zero Length Encode-N our cell

// BWT REVERSE FUNCTIONS - Note that reverse-Huffman is called puff(), not unhuff()

    Bcell& unbwtSarray();
    Bcell& unbwtStree();
    Bcell& unmtf();
    Bcell& unmtf2();
    Bcell& puff();
    Bcell& unarith();
    Bcell& unsrle();
    Bcell& unsrle0();
    Bcell& unsrle0a();
    Bcell& unzle();
    Bcell& unzle1();
    Bcell& unzleWheeler();
    Bcell& unzleWheelerN(int n);

private:

// ZLE

    static void encodeLength        (Gcell<unsigned char>& g, size_t length);
    static void encodeLength1       (Gcell<unsigned char>& g, size_t length);
    static void encodeLengthWheeler (list<unsigned char>& g, size_t length);
    static void encodeLengthWheelerN(Gcell<unsigned char>& g, size_t length, int n);

// BWT SUFFIX ARRAY

    Mcell<size_t>                           // Perform the Manber-Myers algorithm
        makeSortedSuffixArray() const;

// BWT SUFFIX TREE

    class SuffixNode;

    struct Transition {
        short        letter;
        long         left;
        long         right;
        SuffixNode * child;
        Transition(short letter, long left, long right, SuffixNode *child) 
            : letter(letter), left(left), right(right), child(child) { }
    };

    class TransitionList {
    public:
        TransitionList() { ctor++; }
        void insertTransition(const Transition& t);
        bool hasChildren() const { haschild++; return !lst.empty(); }
        bool doesTransitionExist(short a) const;
        Transition getFirstChild() const;
        void removeFirstTransition();
        Transition findTransition(short a) const;
        ~TransitionList() { dtor++; }

        static void displayUsage();

    private:
        list<Transition> lst;

        static size_t ctor;
        static size_t insert;
        static size_t haschild;
        static size_t doesexist;
        static size_t getfirst;
        static size_t removefirst;
        static size_t find;
        static size_t dtor;
    };

    class SuffixNode {
    public:
        TransitionList children;
        SuffixNode *   link;
    
        SuffixNode() : link(NULL) { count++; }
        ~SuffixNode() { count--; }
        bool haschildren() const;
        static void dfs(vector<size_t>& v, SuffixNode * root);

        static void displayInstances() { cout << "Number of SuffixNodes instantiated: " << count << endl; }
    private:
        static size_t count;
    };

    class SuffixTree {
    public:
        SuffixTree(const Bcell& b);
        Mcell<unsigned char> getTransformed() const;

    private:
        SuffixNode *         root;
        SuffixNode           bottom;
        Mcell<short>         text;        // The text being operated on
        const long           infinity;    // The highest valid index in the text,
                                          // the location of the sentinel
                                          // Must equal text.getSize() - 1;
        Mcell<unsigned char> transformed; // The BWT-transformed text

        pair<SuffixNode *, long>   update(pair<SuffixNode *, long> mypair, long i);
        pair<SuffixNode *, long>   canonize(SuffixNode * s, long k, long p);
        pair<bool, SuffixNode *>   test_and_split(SuffixNode * s, long k, long p, short t);
        short                      sanitizetext(long x);
    };
 
// ZLIB

#ifdef ZIP

    static int  myCompress(unsigned char *dest, // Modified compress2()
                unsigned long int *destlen,     // function from zlib 
                const unsigned char *source,    // that calls deflateinit2().
                unsigned long int sourcelen,    // It has the power of
                int num_strategy);              // init2 with the
                                                // convenience of compress.

    Bcell   compressStrategy(int num_strategy) const;

#endif

// SHA-1

    static unsigned long int SHA1Function(
        unsigned long int b,
        unsigned long int c, 
        unsigned long int d, 
        int t);

    static unsigned long int SHA1Constant(int t);

    Bcell SHA1Pad() const;                      // Return a padded version of
                                                // our cell in preparation
                                                // for a SHA-1 hash

    static unsigned long int                    // Circular Left Shift
        circularLeftShift(                      // used in SHA-1
        unsigned long int ul, int n);

// HUFFMAN

    Mcell<Gcell<bool> > HuffmanCodes() const;   // Calc Huffman codes given lengths
    Bcell HuffmanCodeLengths() const;           // Calculate Huffman codelengths

    class Tree {

    public:
        size_t  freq;
        int     byte;        
        Tree *  left;
        Tree *  right;
    private:
        Tree *  up;

        const Tree * findByte(int b) const;     // Find where a given byte is

    public:

        Tree();                                 // Constructor: Makes an empty tree
        void set(Tree & l, Tree & r);           // Creates two trees from one
        void set(size_t f, int b);              // Make our tree into a one-node tree
        int getHeight(int b) const;             // Get the level at which a given byte is at
        ~Tree();                                // Destructor
    };

    void treefill(Tree & t,                     // Fill our cell (of length
        unsigned char h = 0) ;                  // 256) with the codelengths
                                                // of a tree
// ARITH

    class ArithModel {
    public:
        ArithModel();
        const size_t& operator[](size_t i) const;
        void update(int sym);
        ~ArithModel();

    private:
        vector<size_t> freq;
        vector<size_t> cfreq;
    };

    class ArithEncoder {

    private:
        bool finalized;
        list<bool> g;
        unsigned long low;
        unsigned long high;
        unsigned long fbits;
        ArithModel acm;

        void bitPlusFollow(bool bit);

    public:
        ArithEncoder();
        void encodeSymbol(int sym);
        Mcell<bool> finalize();
        ~ArithEncoder();
    };

    class ArithDecoder {

    private:
        Bcell *pcell;
        size_t loc;
        unsigned long value;
        unsigned long low;
        unsigned long high;
        ArithModel acm;

        bool inputBit();

    public:
        ArithDecoder(Bcell *pc);
        int decodeSymbol();
        ~ArithDecoder();
    };

// PNG

    static unsigned char samplesPerPixel(unsigned char colortype);

    static unsigned long bytesPerUnfilteredScanLine(
        unsigned long width, unsigned char bitdepth, unsigned char colortype);

    static unsigned long bytesPerFilteredScanLine(
        unsigned long width, unsigned char bitdepth, unsigned char colortype);

    static unsigned long decompressedSize(
        unsigned long width, unsigned long height,
        unsigned char bitdepth, unsigned char colortype);

    static void checkBitdepthAndColortype(
        unsigned char bitdepth, unsigned char colortype);

    static unsigned char paethPredictor(
        unsigned char a, unsigned char b, unsigned char c);

    static unsigned char bpp(
        unsigned char bitdepth, unsigned char colortype);
};

#endif // End of: #ifndef BCELL_HH
