// bcell_bwt_sarray.cc - Memory Cell Abstraction Interface (Bcell: Byte Mcell)
// Burrows-Wheeler Transformation functions, Suffix Array algorithm
// GPLed - Stephan T. Lavavej - http://stl.caltech.edu

#include "bcell.hh"

Bcell& Bcell::bwtCompressSarray(bool verbose) {

    Bcell sig;
    sig = BWTZIP_SARRAY_SIG;
    Bcell id("Compression Method ID", 1);
    id[0] = 255;
    
    if (size <= MIN_LENGTH_TO_COMPRESS) {
        id[0] = 0;
        *this = sig + id + *this;
        return *this;
    }

    Mcell<long double> timings("Timing Information", 4);

    clockStart(verbose);

    bwtSarray();    timings[0] = clockReport();
    mtf2();         timings[1] = clockReport();
    srle0a();       timings[2] = clockReport();
    arith();        timings[3] = clockReport();

    if (verbose) {
        cout << "BWT Sarray: " << timings[0] << endl;
        cout << "MTF-2:      " << timings[1] << endl;
        cout << "SRLE-0-A:   " << timings[2] << endl;
        cout << "Arith:      " << timings[3] << endl;
    }

    id[0] = 1;
    *this = sig + id + *this;

    return *this;
}

Bcell& Bcell::bwtDecompressSarray(bool verbose) {
    if (size < 8 || readULL(0) != BWTZIP_SARRAY_SIG) {
        cout << "Not compressed with bwtzip/sarray!" << endl;
        exit(EXIT_FAILURE);
    }

    unsigned char id = (*this)[8];
    truncate(9, size);
    
    if (id == 0) { // Not compressed
        return *this;
    }

    clockStart(verbose);
    Mcell<long double> timings("Timing Information", 4);
    
    unarith();      timings[0] = clockReport();
    unsrle0a();     timings[1] = clockReport();
    unmtf2();       timings[2] = clockReport();
    unbwtSarray();  timings[3] = clockReport();
    
    if (verbose) {
        cout << "Unarith:      " << timings[0] << endl;
        cout << "UnSRLE-0-A:   " << timings[1] << endl;
        cout << "UnMTF-2:      " << timings[2] << endl;
        cout << "UnBWT Sarray: " << timings[3] << endl;
    }

    return *this;
}

// Burrows-Wheeler Transform our cell with the Suffix Array algorithm
Bcell& Bcell::bwtSarray() {
    if (size < 2) {
        cout << "ERROR: Bcell of length less than 2 passed to bwtSarray()." << endl;
        exit(EXIT_FAILURE);
    }

    Mcell<size_t> circstr2("BWT Substring Array DoubleLength", size * 2);

    Bcell doubled = *this + *this;

    circstr2 = doubled.makeSortedSuffixArray();

    Mcell<size_t> circstr("BWT Substring Array", size);

    size_t index = 0;
    size_t index2 = 0;

    while (index < size) {
        if (circstr2[index2] >= size) {
            index2++;
        } else {
            circstr[index++] = circstr2[index2++];
        }
    }

    // Find primary index

    Bcell primaryindex;

    for (size_t i = 0; i < size; i++) {
        if (circstr[i] == 1) {
            primaryindex = (unsigned long long) i;
            break;
        }
    }

    Bcell bwtoutput("BWT Output", size);

    for (size_t i = 0; i < size; i++) {
        bwtoutput[i] = (*this)[(circstr[i] + size - 1) % size];
    }

    *this = primaryindex + bwtoutput;

    return *this;
}

Mcell<size_t> Bcell::makeSortedSuffixArray() const {
    const size_t n = size;
    Mcell<size_t> pos("Suffix Array", n), prm("Prm", n), count("Count", n);
    Mcell<bool> bh("BH", n + 1), b2h("B2H", n + 1);

    pos.fill(0);
    prm.fill(0);
    count.fill(0);
    bh.fill(false);
    b2h.fill(false);

    // FIRST PHASE BUCKET SORT: Fill in pos

    size_t index = 0;

    for (int a = 0; a < 256; a++) { // For each type of byte
        for (size_t i = 0; i < n; i++) {
            if ((*this)[i] == a) {
                pos[index++] = i;
            }
        }
    }

    // Fill in prm

    for (size_t i = 0; i < n; i++) {
        prm[pos[i]] = i;
    }

    // Fill in bh

    bh[0] = true;
    unsigned char lastchar = (*this)[pos[0]];
    for (size_t i = 1; i < n; i++) {
        if ((*this)[pos[i]] != lastchar) {
            bh[i] = true;
            lastchar = (*this)[pos[i]];
        }
    }
    bh[n] = true;

    // INDUCTIVE SORTING STAGE

    for (unsigned long h = 1; h < n; h *= 2) {
        unsigned long m = 0;

        for (unsigned long i = 0; i < n; i++) {
            if (bh[i]) {
                m = i;
                count[m] = 0;
            }
            prm[pos[i]] = m;            
        }

        unsigned long d = n - h;
        unsigned long e = prm[d];
        b2h[prm[d] = e + count[e]++] = true;

        for (unsigned long i = 0; i < n; i++) {
            if (bh[i]) {    // We are at the start of a bucket
                unsigned long l = i;
                unsigned long r = i;
                for (unsigned long j = i + 1; j < n + 1; j++) {
                    if (bh[j]) {
                        r = j - 1;
                        break;
                    }
                }
                // The bucket is [l, r]
                for (unsigned long cc = l; cc <= r; cc++) {
                    const unsigned long poscc = pos[cc];
                    if (poscc >= h && poscc <= n + h - 1) {
                        unsigned long dd = poscc - h;
                        e = prm[dd];
                        b2h[prm[dd] += count[e]++] = true;
                    }
                }

                for (unsigned long cc = l; cc <= r; cc++) {
                    const unsigned long poscc = pos[cc];
                    if (poscc >= h && poscc <= n + h - 1) {
                        unsigned long dd = poscc - h;
                        const unsigned long prmdd = prm[dd];

                        if (b2h[prmdd]) {
                            for (unsigned long j = prmdd + 1; ; j++) {
                                if (bh[j] || !b2h[j]) {
                                    e = j;
                                    break;
                                }
                            }
                            for (unsigned long f = prmdd + 1; f < e; f++) {
                                b2h[f] = false;
                            }
                        }
                    }
                }
            }
        }

        for (unsigned long i = 0; i < n; i++) {
            pos[prm[i]] = i;
            if (b2h[i]) {
                bh[i] = true;
            }
        }
    }

    return pos;
}

// Reverse Burrows-Wheeler Transform our cell using the Fenwick algorithm; assumes Suffix Array forward transformation
Bcell& Bcell::unbwtSarray() {
    size_t index = readULL(0);                  // Read primary index
    truncate(8, size);                          // Remove encoded primary index
    Bcell dcell("UnBWT Decoded Cell", size);    // Holds original data

    Mcell<size_t> freqs = getFrequencies();     // Fenwick's K
    Mcell<size_t> links("Links", size);         // Fenwick's L
    Mcell<size_t> mapping("Mapping", 256);      // Fenwick's M
    
    mapping[0] = 0;                             // Build M
    for (int j = 1; j < 256; j++) {
        mapping[j] = mapping[j - 1] + freqs[j - 1];
    }
    
    for (size_t i = 0; i < size; i++) {         // Build L
        links[i] = mapping[(*this)[i]]++;
    }

    for (long long int i = size - 1; i >= 0; i--) {         // Perform Reverse-BWT
        index = links[index];
        dcell[i] = (*this)[index];        
    }
    
    *this = dcell;
    return *this;
}
