// bcell_mtf.cc - Memory Cell Abstraction Interface (Bcell: Byte Mcell)
// Move-to-Front functions
// GPLed - Stephan T. Lavavej - http://stl.caltech.edu

#include "bcell.hh"

// The MTF algorithm is a byte-for-byte streaming recoding. It takes an
// input byte and outputs the number of distinct other bytes which have
// been seen since its last occurrence. At the beginning, byte 0x00 is
// assumed to have most recently occurred, with 0x01 second most recent,
// etc.

// In practice, an array of 256 bytes is maintained. (I variously refer
// to this as a list or a stack, though it is not implemented with
// either of those data structures.) In position 0 of this array, the
// most recently seen byte is stored. In position 1, the second most
// recently seen, etc. When a byte is read from the input stream, its
// location is found in the array and that location is emitted. Then
// that byte is removed from that position, everything in lower positions
// is moved down, and the byte is put into position 0.
// (If the byte read was the most recently seen byte, then a 0 is
// emitted and the array is not changed.)
// The reverse transformation follows a similar process (it can easily
// be understood by looking at the code).

// The MTF-2 algorithm operates exactly the same way, except that it
// changes how things are moved in the array. If the byte was at any
// position > 1, then it is moved to position 1. If the byte was at
// position 1 and the last output byte was not 0, then it is moved to
// position 0. Otherwise it is not moved.

// Move-to-Front transform our cell
Bcell& Bcell::mtf() {

    Bcell mtfstack("MTF Stack", 256);

    for (int n = 0; n < 256; n++) {     // Initalize the stack
        mtfstack[n] = n;
    }

    for (size_t st = 0; st < size; st++) {
        unsigned char uc, out;

        uc = (*this)[st];               // Input character
        out = mtfstack.find(uc);        // Where is it in the stack?
        (*this)[st] = out;              // Write that back
        if (out != 0) {                 // If char read isn't 1st
            mtfstack.shiftdown(out, 0); // Move characters down
            mtfstack[0] = uc;           // Put read character 1st
        }
    }
    return *this;
}

// Reverse Move-to-Front transform our cell
Bcell& Bcell::unmtf() {

    Bcell mtfstack("Un-MTF Stack", 256);

    for (int n = 0; n < 256; n++) {     // Initialize the stack
        mtfstack[n] = n;
    }

    for (size_t st = 0; st < size; st++) {
        unsigned char uc, out;

        uc = (*this)[st];               // Input character
        out = mtfstack[uc];             // Decoded character
        (*this)[st] = out;              // Write decoded character
        if (uc != 0) {                  // If value written wasn't 1st
            mtfstack.shiftdown(uc, 0);  // Move characters down
            mtfstack[0] = out;          // Put read character 1st
        }
    }
    return *this;
}

// Move-to-Front transform our cell
Bcell& Bcell::mtf2() {

    Bcell mtfstack("MTF Stack", 256);

    for (int n = 0; n < 256; n++) {     // Initalize the stack
        mtfstack[n] = n;
    }

    unsigned char lastout = 0;

    for (size_t st = 0; st < size; st++) {
        unsigned char uc, out;

        uc = (*this)[st];               // Input character
        out = mtfstack.find(uc);        // Where is it in the stack?
        (*this)[st] = out;              // Write that back


        if (out > 1) {                      // If char read is from a high position
            mtfstack.shiftdown(out - 1, 1); // Move characters down
            mtfstack[1] = uc;               // Put read character second
        } else if (out == 1 && lastout != 0) {
            mtfstack[1] = mtfstack[0];
            mtfstack[0] = uc;               // Put read character first
        }                                   // If out == 0 then do nothing

        lastout = out;                      // Update last output number
    }
    return *this;
}

// Reverse Move-to-Front transform our cell
Bcell& Bcell::unmtf2() {

    Bcell mtfstack("Un-MTF Stack", 256);

    for (int n = 0; n < 256; n++) {         // Initialize the stack
        mtfstack[n] = n;
    }

    unsigned char lastread = 0;

    for (size_t st = 0; st < size; st++) {
        unsigned char uc, out;

        uc = (*this)[st];                   // Input character
        out = mtfstack[uc];                 // Decoded character
        (*this)[st] = out;                  // Write decoded character
        if (uc > 1) {                       // If value written was from a high position
            mtfstack.shiftdown(uc - 1, 1);  // Move characters down
            mtfstack[1] = out;              // Put read character 1st
        } else if (uc == 1 && lastread != 0) {
            mtfstack[1] = mtfstack[0];
            mtfstack[0] = out;
        }                                   // If uc == 0 then do nothing

        lastread = uc;
    }
    return *this;
}
