// bcell.cc - Memory Cell Abstraction Interface (Bcell: Byte Mcell)
// SHA-1 functions
// GPLed - Stephan T. Lavavej - http://stl.caltech.edu

#include "bcell.hh"

// Return the 160-bit, 20-byte SHA-1 hash of our memory cell
Bcell Bcell::SHA1() const {
    Bcell paddedcell = SHA1Pad();
    Bcell hash("SHA-1 Hash", 20);

    unsigned long int a, b, c, d, e;    // First 5-word buffer in Section 7 of the SHA-1 spec
    Mcell<unsigned long> h("Second 5-Word SHA-1 Buffer", 5);
    Mcell<unsigned long> w("80-Word SHA-1 Sequence", 80);
    unsigned long int temp;

    h[0] = 0x67452301UL;
    h[1] = 0xEFCDAB89UL;
    h[2] = 0x98BADCFEUL;
    h[3] = 0x10325476UL;
    h[4] = 0xC3D2E1F0UL;

    for (size_t st = 0; st < paddedcell.size; st += 64) {

        for (int i = 0; i < 16; i++) {              // Begin processing M sub i.
            w[i] = paddedcell.readUL(st + i * 4);   // Step a
        }

        for (int t = 16; t < 80; t++) {
            w[t] = circularLeftShift(
                w[t - 3] ^ w[t - 8] ^ w[t - 14] ^ w[t - 16], 1); // Step b
        }

        a = h[0];   // Step c
        b = h[1];
        c = h[2];
        d = h[3];
        e = h[4];

        for (int t = 0; t < 80; t++) {
            temp = circularLeftShift(a, 5) + SHA1Function(b, c, d, t) + e + w[t] + SHA1Constant(t);
            e = d;
            d = c;
            c = circularLeftShift(b, 30);
            b = a;
            a = temp;
        }

        h[0] += a;
        h[1] += b;
        h[2] += c;
        h[3] += d;
        h[4] += e;
    }

    for (int i = 0; i < 5; i++) {
        for (int j = 0; j < 4; j++) {
            hash[i * 4 + j]  = decomposeUL(h[i], j);
        }
    }

    return hash;
}

// From Section 5 of the SHA-1 spec
unsigned long int Bcell::SHA1Function(unsigned long int b,
    unsigned long int c, unsigned long int d, int t) {
    
    if ((t < 0) || (t > 79)) {
        cout << "ERROR: Invalid t passed to SHA1Function()." << endl;
        exit(EXIT_FAILURE);
    }

    if ((0 <= t) && (t <= 19)) {
        return (b & c) | (~b & d);
    } else if ((20 <= t) && (t <= 39)) {
        return b ^ c ^ d;
    } else if ((40 <= t) && (t <= 59)) {
        return (b & c) | (b & d) | (c & d);
    } else if ((60 <= t) && (t <= 79)) {
        return b ^ c ^ d;
    } else {
        cout << "IMPOSSIBLE!" << endl;
        exit(EXIT_FAILURE);
    }
}

// From Section 6 of the SHA-1 spec
unsigned long int Bcell::SHA1Constant(int t) {
    if ((t < 0) || (t > 79)) {
        cout << "ERROR: Invalid t passed to SHA1Constant()." << endl;
        exit(EXIT_FAILURE);
    }

    if ((0 <= t) && (t <= 19)) {
        return 0x5A827999UL;
    } else if ((20 <= t) && (t <= 39)) {
        return 0x6ED9EBA1UL;
    } else if ((40 <= t) && (t <= 59)) {
        return 0x8F1BBCDCUL;
    } else if ((60 <= t) && (t <= 79)) {
        return 0xCA62C1D6UL;
    } else {
        cout << "IMPOSSIBLE!" << endl;
        exit(EXIT_FAILURE);
    }
}

// PURPOSE: Return a padded version of our cell in preparation for a SHA-1 hash
Bcell Bcell::SHA1Pad() const {

    // SHA-1 padding is simple.
    // First, a '1' bit is appended to the data.
    // Then, '0' bits are appended until the length of the data in bits mod 512 is 448.
    // Finally, the length of the original message is encoded as a 64-bit big-endian
    // integer and appended.

    // In byte terms:
    // First, 0x80 is appended to the data.
    // Then, 0x00 bytes are appended until the length of the data mod 64 is 56.
    // Finally, the length is appended, using 8 bytes.

    int n;                  // The number of bytes to append
    int s = size % 64;      // The size of the data mod 64

    n = 1; // 0x80 byte
    s++;

    // 0x00 bytes
    while (s != 56) {
        n++;
        ++s %= 64;
    }

    // 64-bit length
    n += 8;

    Bcell padding("SHA-1 Padding", n);

    padding.fill(0x00);

    padding[0] = 0x80;

    unsigned long long int ullsize = size;
    ullsize *= 8;

    for (int i = 0; i < 8; i++) {
        padding[i + n - 8] = decomposeULL(ullsize, i);
    }

    return *this + padding;
}
