// bcell_zle.cc - Memory Cell Abstraction Interface (Bcell: Byte Mcell)
// Zero Length Encoder functions
// GPLed - Stephan T. Lavavej - http://stl.caltech.edu

#include "bcell.hh"

// The Zero Length Encoder algorithm

// Cleartext: A run of N 0x00 bytes
// Encoded as: N is written down as a big-endian binary number.
// 0 bits become 0a symbols and 1 bits become 0b symbols.
// 0a is encoded as 0x00 and 0b is encoded as 0x01.

// Cleartext: A byte B = 0x01 or 0x02 or ... or 0xFD
// Encoded as: B + 1

// Cleartext: A 0xFE byte
// Encoded as: 0xFF 0x00

// Cleartext: A 0xFF byte
// Encoded as: 0xFF 0x01

Bcell& Bcell::zle() {
    Gcell<unsigned char> g;

    size_t length = 0;

    for (size_t i = 0; i < size; i++) {
        const unsigned char byte = (*this)[i];

        if (byte == 0x00) {
            length++;
        } else if (length != 0) {
            encodeLength(g, length);
            length = 0;
        }

        if (byte == 0xFE) {
            g.appendData(0xFF);
            g.appendData(0x00);
        } else if (byte == 0xFF) {
            g.appendData(0xFF);
            g.appendData(0x01);
        } else if (byte != 0x00) {
            g.appendData(byte + 1);
        }
    }

    if (length != 0) {
        encodeLength(g, length);
    }

    *this = g.getMcell();
    return *this;
}

void Bcell::encodeLength(Gcell<unsigned char>& g, size_t length) {
    if (length <= 0) {
        cout << "Bcell::encodeLength() given non-positive length!" << endl;
        exit(EXIT_FAILURE);
    }

    Gcell<unsigned char> temp;

    while (length != 0) {
        temp.prependData(length & 1);
        length >>= 1;
    }

    g.destructiveAppend(temp);
}

Bcell& Bcell::unzle() {
    Gcell<unsigned char> g;

    size_t length = 0;

    for (size_t i = 0; i < size; i++) {
        const unsigned char byte = (*this)[i];

        if (byte <= 0x01) {
            length <<= 1;
            length += byte;
        } else if (length != 0) {
            while (length != 0) {
                g.appendData(0x00);
                length--;
            }
        }
        
        if (byte == 0xFF) {
            const unsigned char nextbyte = (*this)[++i];

            if (nextbyte == 0x00) {
                g.appendData(0xFE);
            } else if (nextbyte == 0x01) {
                g.appendData(0xFF);
            } else {
                cout << "Problem in Bcell::unzle()!" << endl;
                exit(EXIT_FAILURE);
            }
        } else if (byte > 0x01) {
            g.appendData(byte - 1);
        }
    }

    if (length != 0) {
        while (length != 0) {
            g.appendData(0x00);
            length--;
        }
    }

    *this = g.getMcell();
    return *this;
}

// The Zero Length Encoder-1 algorithm

// ZLE-1 operates the same as ZLE, except that the length of a run
// is decremented before it is encoded. A run of 1 zero is encoded as a run of 0 zeros.
// A run of 2 zeros is encoded as a run of 1 zero, and so forth. This is because
// an actual run of 0 zeros can't occur. Thus, it saves a bit when expressing some
// lengths.

Bcell& Bcell::zle1() {
    Gcell<unsigned char> g;

    size_t length = 0;

    for (size_t i = 0; i < size; i++) {
        const unsigned char byte = (*this)[i];

        if (byte == 0x00) {
            length++;
        } else if (length != 0) {
            encodeLength1(g, length);
            length = 0;
        }

        if (byte == 0xFE) {
            g.appendData(0xFF);
            g.appendData(0x00);
        } else if (byte == 0xFF) {
            g.appendData(0xFF);
            g.appendData(0x01);
        } else if (byte != 0x00) {
            g.appendData(byte + 1);
        }
    }

    if (length != 0) {
        encodeLength1(g, length);
    }

    *this = g.getMcell();
    return *this;
}

void Bcell::encodeLength1(Gcell<unsigned char>& g, size_t length) {
    if (length <= 0) {
        cout << "Bcell::encodeLength1() given non-positive length!" << endl;
        exit(EXIT_FAILURE);
    }

    if (--length == 0) {
        g.appendData(0x00);
    } else {
        Gcell<unsigned char> temp;

        while (length != 0) {
            temp.prependData(length & 1);
            length >>= 1;
        }

        g.destructiveAppend(temp);
    }
}

Bcell& Bcell::unzle1() {
    Gcell<unsigned char> g;

    size_t length = 0;
    bool   run = false;

    for (size_t i = 0; i < size; i++) {
        const unsigned char byte = (*this)[i];

        if (byte <= 0x01) {
            run = true;
            length <<= 1;
            length += byte;
        } else if (run) {
            run = false;
            length++;   // Increment length
    
            while (length != 0) {
                g.appendData(0x00);
                length--;
            }
        }
        
        if (byte == 0xFF) {
            const unsigned char nextbyte = (*this)[++i];

            if (nextbyte == 0x00) {
                g.appendData(0xFE);
            } else if (nextbyte == 0x01) {
                g.appendData(0xFF);
            } else {
                cout << "Problem in Bcell::unzle()!" << endl;
                exit(EXIT_FAILURE);
            }
        } else if (byte > 0x01) {
            g.appendData(byte - 1);
        }
    }

    if (run) {
        length++;   // Increment length

        while (length != 0) {
            g.appendData(0x00);
            length--;
        }
    }

    *this = g.getMcell();
    return *this;
}

// The Wheeler Zero Length Encoder algorithm

// D.J. Wheeler's Zero Length Encoder is even more clever than ZLE and ZLE-1. It operates
// identically to ZLE, except in the way it treats the length of runs.
// When encoding the length of a run, first it is INCREMENTED. The possible run lengths
// 1, 2, 3, 4, 5, 6, 7, 8, 9, ...
// are incremented to:
// 2, 3, 4, 5, 6, 7, 8, 9, 10, ...
// Then they are written in binary:
// 10, 11, 100, 101, 110, 111, 1000, 1001, 1010, ...
// Note that the most significant bit of all of these numbers is 1. This bit is dropped.
// 0, 1, 00, 01, 10, 11, 000, 001, 010, ...
// Then these are encoded as in ZLE.

Bcell& Bcell::zleWheeler() {
    list<unsigned char> g;

    size_t length = 0;

    for (size_t i = 0; i < size; i++) {
        const unsigned char byte = (*this)[i];

        if (byte == 0x00) {
            length++;
        } else if (length != 0) {
            encodeLengthWheeler(g, length);
            length = 0;
        }

        if (byte == 0xFE) {
            g.push_back(0xFF);
            g.push_back(0x00);
        } else if (byte == 0xFF) {
            g.push_back(0xFF);
            g.push_back(0x01);
        } else if (byte != 0x00) {
            g.push_back(byte + 1);
        }
    }

    if (length != 0) {
        encodeLengthWheeler(g, length);
    }

    *this = g;
    return *this;
}

void Bcell::encodeLengthWheeler(list<unsigned char>& g, size_t length) {
    if (length <= 0) {
        cout << "Bcell::encodeLengthWheeler() given non-positive length!" << endl;
        exit(EXIT_FAILURE);
    }

    length++;                           // Increment length first

    list<unsigned char> temp;

    while (length != 1) {               // Do not write most significant bit
        temp.push_front(length & 1);
        length >>= 1;
    }

    g.splice(g.end(), temp);
}

Bcell& Bcell::unzleWheeler() {
    list<unsigned char> g;

    size_t length = 0;
    bool   run = false;

    for (size_t i = 0; i < size; i++) {
        const unsigned char byte = (*this)[i];

        if (byte <= 0x01) {
            if (run == false) { // We must ghostwrite the most significant bit to be 1
                length = 1;
            }

            run = true;
            length <<= 1;
            length += byte;
        } else if (run) {
            run = false;
            length--;   // Decrement length to undo the effect of the Wheeler ZLE
    
            while (length != 0) {
                g.push_back(0x00);
                length--;
            }
        }
        
        if (byte == 0xFF) {
            const unsigned char nextbyte = (*this)[++i];

            if (nextbyte == 0x00) {
                g.push_back(0xFE);
            } else if (nextbyte == 0x01) {
                g.push_back(0xFF);
            } else {
                cout << "Problem in Bcell::unzle()!" << endl;
                exit(EXIT_FAILURE);
            }
        } else if (byte > 0x01) {
            g.push_back(byte - 1);
        }
    }

    if (run) {
        length--;

        while (length != 0) {
            g.push_back(0x00);
            length--;
        }
    }

    *this = g;
    return *this;
}

// The Wheeler Zero Length Encoder-N algorithm

// This is identical to the Wheeler ZLE algorithm, except that runs are expressed in
// base N.

Bcell& Bcell::zleWheelerN(int n) {
    if (n < 2 || n > 255) {
        cout << "Invalid base passed to zleWheelerN()!" << endl;
        exit(EXIT_FAILURE);
    }

    Gcell<unsigned char> g;

    size_t length = 0;

    for (size_t i = 0; i < size; i++) {
        const unsigned char byte = (*this)[i];

        if (byte == 0x00) {
            length++;
        } else if (length != 0) {
            encodeLengthWheelerN(g, length, n);
            length = 0;
        }

        if (byte >= (0xFF - n + 1)) {
            g.appendData(0xFF);
            g.appendData(byte - (0xFF - n + 1));
        } else if (byte != 0x00) {
            g.appendData(byte - 1 + n);
        }
    }

    if (length != 0) {
        encodeLengthWheelerN(g, length, n);
    }

    *this = g.getMcell();
    return *this;
}

void Bcell::encodeLengthWheelerN(Gcell<unsigned char>& g, size_t length, int n) {
    if (length <= 0) {
        cout << "Bcell::encodeLengthWheeler() given non-positive length!" << endl;
        exit(EXIT_FAILURE);
    }
    if (n < 2 || n > 255) {
        cout << "Invalid base passed to encodeLengthWheelerN()!" << endl;
        exit(EXIT_FAILURE);
    }

    Gcell<unsigned char> temp;

    while (length != 0) {
        if (length % n == 0) {
            temp.prependData(n - 1);
            length -= n;
        } else {
            temp.prependData(length % n - 1);
        }
           
        length /= n;
    }

    g.destructiveAppend(temp);
}

Bcell& Bcell::unzleWheelerN(int n) {
    if (n < 2 || n > 255) {
        cout << "Invalid base passed to unzleWheelerN()!" << endl;
        exit(EXIT_FAILURE);
    }
    
    Gcell<unsigned char> g;

    size_t length = 0;

    for (size_t i = 0; i < size; i++) {
        const unsigned char byte = (*this)[i];

        if (byte < n) {
            length *= n;
            length += byte + 1;
        } else if (length != 0) {
            while (length != 0) {
                g.appendData(0x00);
                length--;
            }
        }
        
        if (byte == 0xFF) {
            g.appendData(0xFF - n + 1 + (*this)[++i]);
        } else if (byte >= n) {
            g.appendData(byte - n + 1);
        }
    }

    if (length != 0) {
        while (length != 0) {
            g.appendData(0x00);
            length--;
        }
    }

    *this = g.getMcell();
    return *this;
}
